-------------------------------------------------------------------------
--  CONSTRUCAO DA ARQUITETURA CLEOPATRA COMPLETA
--
--  Fernando Moraes
--
--  Inicio: 11/maio/1998             Ultima modificacao:  31/outubro/2001
--
--  15/10/2000 - simplificacao da Unidade de Controle
--  15/10/2000- correcao dos jumps indiretos
--
--  OBSERVACAO:  nao tem o flag do overflow
--
--------------------------------------------------------------------------
 
--------------------------------------------------------------------------
--
-- package com tipos basicos, e funcoes de soma e comparacao com zero
--
--------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;

package cleo is
    
    subtype opcode  is std_logic_vector(2 downto 0);
    subtype regsize is std_logic_vector(7 downto 0);
    type ram is array (0 to 255) of regsize;
    
    type microinstrucao is record 
               w,r,u: opcode;
               ce, rw, nz, c : std_logic;
    end record;
    
    -- CODIFICACAO DAS INSTRUCOES
    constant nota    : regsize:=  x"00";
    constant stadir  : regsize:=  x"24";    
    constant staind  : regsize:=  x"28";
    constant starel  : regsize:=  x"2C";    
    constant ldaim   : regsize:=  x"40";  
    constant ldadir  : regsize:=  x"44";
    constant ldaind  : regsize:=  x"48";
    constant ldarel  : regsize:=  x"4C";
    constant addim   : regsize:=  x"50";     
    constant adddir  : regsize:=  x"54";     
    constant addind  : regsize:=  x"58";
    constant addrel  : regsize:=  x"5C";
    constant orim    : regsize:=  x"60";
    constant ordir   : regsize:=  x"64";
    constant orind   : regsize:=  x"68";
    constant orrel   : regsize:=  x"6C";
    constant andim   : regsize:=  x"70";
    constant anddir  : regsize:=  x"74";
    constant andind  : regsize:=  x"78";
    constant andrel  : regsize:=  x"7C";
    constant jmpdir  : regsize:=  x"84";
    constant jmpind  : regsize:=  x"88";
    constant jmprel  : regsize:=  x"8C";
    constant jcdir   : regsize:=  x"94";
    constant jcind   : regsize:=  x"98";
    constant jcrel   : regsize:=  x"9C";
    constant jndir   : regsize:=  x"A4";
    constant jnind   : regsize:=  x"A8";
    constant jnrel   : regsize:=  x"AC";
    constant jzdir   : regsize:=  x"B4";
    constant jzind   : regsize:=  x"B8";
    constant jzrel   : regsize:=  x"BC";
    constant jsrdir  : regsize:=  x"C4";
    constant jsrind  : regsize:=  x"C8";
    constant jsrrel  : regsize:=  x"CC";
    constant rts     : regsize:=  x"D0";
    constant hlt     : regsize:=  x"E0";
   
    procedure somaAB  (  signal A,B: in  regsize;   signal Cin: in STD_LOGIC;
                         signal S:   out regsize;   signal Cout:out STD_LOGIC);
    function is_zero  (  signal A:  in  regsize  ) return std_logic ;
end cleo;

package body cleo is

      procedure somaAB  (  signal A,B:  in  regsize;  signal Cin:  in STD_LOGIC;
                         signal S:    out regsize;  signal Cout: out STD_LOGIC) is             
         variable carry : STD_LOGIC;
      begin
           for w in 0 to 7 loop
                if w=0 then carry:=Cin;  end if;
                S(w) <= A(w) xor B(w) xor carry;
                carry := (A(w) and B(w)) or (A(w) and carry) or (B(w) and carry);
           end loop;
           Cout <= carry;
      end somaAB;
    
      function is_zero  (  signal A:  in  regsize  ) return std_logic is           
      begin    
         if A="00000000" then return '1'; else return '0';  end if;
      end is_zero;
           
end cleo;

-------------------------------------------------------------------------
--- registrador de 8 bits, com reset assincrono e enable para escrita
-------------------------------------------------------------------------
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity reg8clear is
    port( clock,reset,ce:in std_logic;   D:in regsize;  Q:out regsize );
end reg8clear;

architecture reg8clear of reg8clear is
begin
  process (clock, reset)
    begin
    if reset='1' then
         Q <= (others=>'0');
    elsif (clock'event and clock='0')then
        if ce='1' then  Q <= D;  end if;
    end if;
  end process;
end reg8clear;

--*********************************************************************--
--*********************************************************************--
---  Descricao estrutural do bloco de dados
--*********************************************************************--
--*********************************************************************--
library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity datapath is
    port( uins          :  in microinstrucao;
          ck, reset     :  in std_logic;
          ir, address   :  out regsize;
          datamem       :  inout regsize;
          n, z, c       :  out std_logic
        ); 
end datapath;

architecture datapath of datapath is

   component reg8clear
    port( clock,reset,ce : in std_logic;   D : in regsize; Q : out regsize );
   end component;
   
   signal zero, cout, sel1,sel2,rmdr,rir,rpc,rac,rrs,wmar,wmdr,wir,wpc,wac,wrs : std_logic; 
   signal um, mdr, reg_ir, pc, ac, rs, outmux, busA, busB, out_ula : regsize;
 
begin  

    ir <= reg_ir;    -- instrucao corrente, a ser utilizada no bloco de controle

    sel1 <= uins.ce and uins.rw;
    outmux <= out_ula  when sel1='0' else datamem;
    
    sel2 <= uins.ce and (not uins.rw);
    datamem <= out_ula when sel2='1' else "ZZZZZZZZ";
    
    --
    -- registradores de dados: mar, mdr, ir, pc, ac, rs
    --
    R1: reg8clear port map ( clock=>ck, reset=>reset, ce=>wmar, D=>out_ula, Q=>address);
    R2: reg8clear port map ( clock=>ck, reset=>reset, ce=>wmdr, D=>outmux,  Q=>mdr);
    R3: reg8clear port map ( clock=>ck, reset=>reset, ce=>wir,  D=>out_ula, Q=>reg_ir);
    R4: reg8clear port map ( clock=>ck, reset=>reset, ce=>wpc,  D=>out_ula, Q=>pc);
    R5: reg8clear port map ( clock=>ck, reset=>reset, ce=>wac,  D=>out_ula, Q=>ac);
    R6: reg8clear port map ( clock=>ck, reset=>reset, ce=>wrs,  D=>out_ula, Q=>rs);
    
    --
    -- tristates para jogar o valor dos registradores nos barramentos
    --
    busB <= mdr     when rmdr='1' else "ZZZZZZZZ";
    busB <= reg_ir  when rir='1'  else "ZZZZZZZZ";
    busA <= pc      when rpc='1'  else "ZZZZZZZZ";
    busA <= ac      when rac='1'  else "ZZZZZZZZ";
    busA <= rs      when rrs='1'  else "ZZZZZZZZ";
    
    ---
    --- Cdigos de escrita nos registradores (codificao de escrita):
    --- 0 MAR, 1 MDR, 2 IR, 3 PC, 4 AC, 5 RS, 6 PC/MDR, 7 nenhum reg. alterado
    ---
    wmar <= '1' when uins.w="000"                 else '0';
    wmdr <= '1' when uins.w="001" or uins.w="110" else '0';
    wir  <= '1' when uins.w="010"                 else '0';
    wpc  <= '1' when uins.w="011" or uins.w="110" else '0';
    wac  <= '1' when uins.w="100"                 else '0';
    wrs  <= '1' when uins.w="101"                 else '0';
    
    ---
    --- Cdigos de leitura nos registradores (codificao de leitura):
    --- 1 MDR, 2 IR, 3 PC, 4 AC, 5 RS, 6 MDR/AC, 7 MDR/PC
    ---
    rmdr <= '1' when uins.r="001" or uins.r="110" or uins.r="111" else '0';
    rir  <= '1' when uins.r="010"                                 else '0'; 
    rpc  <= '1' when uins.r="011" or uins.r="111"                 else '0';
    rac  <= '1' when uins.r="100" or uins.r="110"                 else '0';  
    rrs  <= '1' when uins.r="101"                                 else '0';                                               
   
    ---
    ---  Cdigos da ULA:  
    ---  0 soma, 1 inc, 2 not, 4 passaB, 5 ou, 6 and, 7 passaA
    ---
    um <= "00000001";
    zero <= '0';
    process(uins,busA,busB,um,zero)
     begin
       case uins.u is
        when "000" => somaAB( busA, busB, zero, out_ula, cout);
        when "001" => somaAB( busA, um,   zero, out_ula, cout);
        when "010" => out_ula <= not busA;
        when "100" => out_ula <= busB;
        when "101" => out_ula <= busA or busB;
        when "110" => out_ula <= busA and busB;
        when "111" => out_ula <= busA;
        when others => null;
       end case;
     end process;
    
    ---------------------------------------------------------------------
    --  Processos que determinam as flags de estado                    --
    ---------------------------------------------------------------------
    process(ck,reset,uins)
      begin
      if (reset='1') then
          c <= '0';    
          n <= '0';
          z <= '0';
      elsif ck'event and ck='0' then  
          
          if uins.c='1' then  
                c <= cout;   
          end if;
          
          if uins.nz='1' then
                n <= out_ula(7);
                z <= is_zero(out_ula);
          end if;  
          
      end if;
    end process;

 end datapath;

--*********************************************************************--
--*********************************************************************--
---  Descricao comportamental do bloco de controle
--*********************************************************************--
--*********************************************************************--

library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity control is
    port( ir :   in regsize;
          reset, n, z, c, ck: in std_logic;
          uins : out microinstrucao
        ); 
end control;

architecture control of control is 

  signal jc,jn,jz: std_logic;  
  
  ---
  --- definicao das microinstrucoes em funcao dos registrador destino/origem
  ---
  constant mar_pc    : microinstrucao := ("000", "011", "111",'0','0','0','0'); -- mar <- pc  
  constant mar_mdr   : microinstrucao := ("000", "001", "100",'0','0','0','0'); -- mar <- mdr
  constant mdr_MmarP : microinstrucao := ("110", "011", "001",'1','1','0','0'); -- mdr <- pmem(mar); pc++
  constant mdr_Mmar  : microinstrucao := ("001", "000", "011",'1','1','0','0'); -- mdr <- pmem(mar)
  constant mar_mdrpc : microinstrucao := ("000", "111", "000",'1','1','0','0'); -- mar <- mdr+PC 
  constant ir_mdr    : microinstrucao := ("010", "001", "100",'0','0','0','0'); -- ir <- mdr
  constant pc_mdr    : microinstrucao := ("011", "001", "100",'0','0','0','0'); -- pc <- mdr
  constant pc_mdrpc  : microinstrucao := ("011", "111", "000",'0','0','0','0'); -- pc< mdr + pc
  constant pc_rts    : microinstrucao := ("011", "101", "111",'0','0','0','0'); -- pc <- rts
  constant ac_ac     : microinstrucao := ("100", "100", "010",'0','0','1','0'); -- ac <- not ac ; lnz
  constant rts_pc    : microinstrucao := ("101", "011", "111",'0','0','0','0'); -- rts <- pc 
  constant Mmar_ac   : microinstrucao := ("111", "100", "111",'1','0','0','0'); -- pmem(mar) <- ac    
  constant nop       : microinstrucao := ("111", "000", "000",'0','0','0','0'); -- nao faz nada 
 
  procedure sel_op (signal ir: in std_logic_vector(3 downto 0); signal uins : out microinstrucao ) is
  begin
      case ir is
        when x"4" => uins <= ("100", "001", "100", '0','0', '1','0');   -- ac <- mdr
        when x"5" => uins <= ("100", "110", "000", '0','0', '1','1');   -- ac <- ac + mdr
        when x"6" => uins <= ("100", "110", "101", '0','0', '1','0');   -- ac <- ac or mdr
        when x"7" => uins <= ("100", "110", "110", '0','0', '1','0');   -- ac <- ac and mdr
        when others => null;    
     end case;
  end sel_op;
   
begin
 
  jc <= (ir(7) and (not(ir(6))) and (not(ir(5))) and ir(4));
  jn <= (ir(7) and (not(ir(6))) and ir(5) and (not(ir(4))));
  jz <= (ir(7) and (not ir(6))) and ir(5) and ir(4);

  process  
         
   begin
           
    -- inicializa o acesso a memoria no reset      
    if reset='1' then uins.rw <= '0'; uins.ce <= '0'; end if;
        
    wait until  ck'event and ck='1';
        
    -- fetch --
    uins <=  mar_pc;        wait until  ck'event and ck='1';
    uins <= mdr_MmarP;      wait until  ck'event and ck='1';
    uins <= ir_mdr;         wait until  ck'event and ck='1';
                        
    case ir is
    
      when ldaim | andim | orim | addim => -- LDA, AND, OR, ADD imediato  
        uins <= mar_pc;        wait until  ck'event and ck='1';
        uins <= mdr_MmarP;     wait until  ck'event and ck='1';
        sel_op (ir(7 downto 4),uins);

      when adddir | ldadir | anddir | ordir => -- LDA, AND, OR, ADD direto 
        uins <= mar_pc;        wait until  ck'event and ck='1';
        uins <= mdr_MmarP;     wait until  ck'event and ck='1';
        uins <= mar_mdr;       wait until  ck'event and ck='1';
        uins <= mdr_Mmar;      wait until  ck'event and ck='1';
        sel_op (ir(7 downto 4),uins);
           
      when andind | orind | addind | ldaind =>  -- LDA, AND, OR, ADD imediato 
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= mar_mdr;      wait until  ck'event and ck='1';
         uins <= mdr_Mmar;     wait until  ck'event and ck='1';
         uins <= mar_mdr;      wait until  ck'event and ck='1';
         uins <= mdr_Mmar;     wait until  ck'event and ck='1';
         sel_op (ir(7 downto 4),uins);
       
      when ldarel | addrel | andrel | orrel =>  -- LDA, AND, OR, ADD relativo
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= mar_mdrpc;    wait until  ck'event and ck='1';
         uins <= mdr_Mmar;     wait until  ck'event and ck='1';
         sel_op (ir(7 downto 4),uins);

      when stadir =>   -- STORE direto
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= mar_mdr;      wait until  ck'event and ck='1';
         uins <= Mmar_ac;   -- PMEM(mar) <- mdr    

      when staind =>  -- STORE indireto
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= mar_mdr;      wait until  ck'event and ck='1';
         uins <= mdr_Mmar;     wait until  ck'event and ck='1';  
         uins <= mar_mdr;      wait until  ck'event and ck='1'; 
         uins <= Mmar_ac;   -- PMEM(mar) <- mdr    
    
       when starel =>  -- STORE relativo
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= mar_mdrpc;        wait until  ck'event and ck='1';
         uins <= Mmar_ac;  -- PMEM(mar) <- mdr    

     --
     --  SALTOS EM MODO IMEDIATO/DIRETO DE ENDERECAMENTO
     --
      when jmpdir =>   -- JUMP direto 
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= pc_mdr;   -- pc <- mdr       

      when jcdir | jndir | jzdir =>   -- JUMP if c='1'or z='1' or z='1' direto 
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         if(((jc and c)='1') or ((jn and n)='1') or ((jz and z)='1')) then 
                 uins <= pc_mdr;   -- pc <- mdr 
            else uins <= nop;
         end if;

      when jsrdir => -- JSR direto    
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= rts_pc;       wait until  ck'event and ck='1';
         uins <= pc_mdr;   -- pc <- mdr
 
     --
     --  SALTOS EM MODO INDIRETO DE ENDERECAMENTO
     --
      when jmpind =>   -- JUMP indireto 
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= mar_mdr;      wait until  ck'event and ck='1';
         uins <= mdr_Mmar;     wait until  ck'event and ck='1';
         uins <= pc_mdr;   -- pc <-mdr
     
      when jcind | jnind | jzind =>   -- JUMP if c='1'or z='1' or z='1' indireto 
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= mar_mdr;      wait until  ck'event and ck='1';
         uins <= mdr_Mmar;     wait until  ck'event and ck='1';
         if(((jc and c)='1') or ((jn and n)='1') or ((jz and z)='1')) then 
                 uins <= pc_mdr;   -- pc <- mdr
            else uins <= nop;
         end if;

      when jsrind => -- JSR indireto 
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= mar_mdr;      wait until  ck'event and ck='1';
         uins <= mdr_Mmar;     wait until  ck'event and ck='1';
         uins <= rts_pc;       wait until  ck'event and ck='1';
         uins <= pc_mdr;    -- pc <- mdr

     --
     --  SALTOS EM MODO RELATIVO AO PC DE ENDERECAMENTO
     --
     when jmprel => -- JUMP relativo   
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= pc_mdrpc;   -- pc <- pc + mdr  

      when jcrel | jnrel | jzrel =>   -- JUMP if c='1'or z='1' or z='1' relativo 
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         if(((jc and c)='1') or ((jn and n)='1') or ((jz and z)='1')) then 
                 uins <= pc_mdrpc;   -- pc <- pc + mdr
            else uins <= nop;
         end if;

       when jsrrel =>   -- JSR relativo 
         uins <= mar_pc;       wait until  ck'event and ck='1';
         uins <= mdr_MmarP;    wait until  ck'event and ck='1';
         uins <= rts_pc;       wait until  ck'event and ck='1';
         uins <= pc_mdrpc;    -- pc <- mdr
       
      when rts => -- RTS
         uins <= pc_rts; -- pc <- rs      
    
      when nota => -- NOT A
         uins <= ac_ac;  -- ac <- NOT(ac)
     
      when hlt =>
           
         while reset='0' loop
          wait until  ck'event and ck='1';
         end loop;
        
      when others => null ;  
          
    end case;
  end process;  
  
end control;

--*********************************************************************--
--*********************************************************************--
---  CPU: uniao do bloco de dados com o bloco de controle
--*********************************************************************--
--*********************************************************************--

library IEEE;
use IEEE.Std_Logic_1164.all;
use work.cleo.all;

entity cleopatra is
      port( ck, reset: in std_logic;
            ce, rw: out std_logic; 
            address   :  out regsize;
            datamem   :  inout regsize
           ); 
end cleopatra;
    
architecture cleopatra of cleopatra is
                                         
     component datapath                                 
      port( uins : in microinstrucao;
            ck, reset: in std_logic;
            ir, address   :  out regsize;
            datamem   :  inout regsize;
            n, z, c   :  out std_logic
           ); 
     end component;
     
     component control
       port( ir : in regsize;
             reset,n, z, c,ck : in std_logic;
             uins : out microinstrucao
           ); 
      end component;
    
   signal uins : microinstrucao; 
   signal n,z,c : std_logic;
   signal ir : regsize;
   
begin

   ce <= uins.ce;
   rw <= uins.rw;

   dp: datapath port map ( uins=>uins, ck=>ck, reset=>reset, 
                           ir=>ir, address=>address, datamem=>datamem, n=>n, z=>z, c=>c);  
                           
   ctrl: control port map ( ir=>ir,reset=>reset, n=>n, z=>z, c=>c, ck=>ck, uins=>uins);                                 
end cleopatra;

--*********************************************************************--
--*********************************************************************--
---  test bench - oberservar as 3 partes componentes do circuito:
---  (1)  CPU: bloco de dados e bloco de controle
---  (2)  interface com a memoria RAM e o preenchimento da RAM
---  (3)  geracao do clock e reset para o circuito 
--
--   verificado em 03/novembro/1998    -   Fernando Gehm Moraes
--
--  Correcao do dia 13/11: escrita na memoria (os tempos de 1ps sao para
--     permitir a escrita nos sinais de inicializacao - preenchimento da
--     memoria)
--
--
--*********************************************************************--
--*********************************************************************--
library IEEE;
use IEEE.Std_Logic_1164.all;
use IEEE.Std_Logic_unsigned.all;   -- uso conv_integer
use IEEE.Std_Logic_arith.all;
use STD.TEXTIO.all;
use work.cleo.all;

entity tb is
end tb;
        
architecture tb of tb is
                                      
   signal ck, reset, ce, rw, go: std_logic;
   signal address, data : regsize;
   file INFILE : TEXT open READ_MODE is "testa.txt";
   signal memoria : ram; 
   signal ops, endereco : integer;
   
   component cleopatra
     port( ck, reset: in std_logic;
            ce, rw: out std_logic; 
                address   :  out regsize;
                datamem   :  inout regsize
           ); 
   end component;
   
begin
   --
   -- inclui a cleopatra no test_bench
   --
   
   cpu : cleopatra port map(ck=>ck,reset=>reset,ce=>ce,rw=>rw, address=>address,datamem=>data);

   --- le da memoria
   data<=memoria(CONV_INTEGER(address)) when ce='1' and rw='1' else "ZZZZZZZZ";

   -- escreve na memoria, de maneira sincrona, como nos registradores
   --
   --  aqui foi a modificacao do dia 13/11 = ESCRITA NA RAM APENAS EM UM
   --  LUGAR. Se isto nao for feito havera valores indefinidos na simulacao
   --
   --  alterado por Fernando Moraes
   --
   process(go, ce, rw, ck)
    begin
       if go'event and go='1' then
          if endereco>=0 and endereco <= 255 then
            memoria(endereco) <= conv_std_logic_vector(ops,8);
           end if;
       elsif ck'event and ck='0' and ce='1' and rw='0' then
                  if CONV_INTEGER(address)>=0 and CONV_INTEGER(address) <= 255 then
                    memoria(CONV_INTEGER(address)) <= data;
          end if;
       end if;
   end process;
   
   reset <= '1' , '0' after 5ns ;   
         
   --- gera o clock e conta conta ciclos de maquina sao utilizados (debug)                   
   process
            begin
             ck <= '1', '0' after 10ns;
             wait for 20ns;
   end process;

   --- realiza a carga na memoria quando acontece o reset
   process
      variable IN_LINE : LINE;                  -- pointer to string
      variable linha : string(1 to 5);
      begin
       
       wait until reset = '1';

       while NOT( endfile(INFILE)) loop -- end file checking
          readline(INFILE,IN_LINE);             -- read line of a file
          read(IN_LINE, linha); 
           
          case linha(1) is  
           when '0' => endereco <= 0;
           when '1' => endereco <= 1;
           when '2' => endereco <= 2;
           when '3' => endereco <= 3;
           when '4' => endereco <= 4;
           when '5' => endereco <= 5;
           when '6' => endereco <= 6;
           when '7' => endereco <= 7;
           when '8' => endereco <= 8;
           when '9' => endereco <= 9;
           when 'A' => endereco <=10;
           when 'B' => endereco <=11;
           when 'C' => endereco <=12;
           when 'D' => endereco <=13;
           when 'E' => endereco <=14;
           when 'F' => endereco <=15;
           when others => null;
          end case;
     
          wait for 1 ps;
          
          case linha(2) is  
           when '0' => endereco <= endereco*16 + 0;
           when '1' => endereco <= endereco*16 + 1;
           when '2' => endereco <= endereco*16 + 2;
           when '3' => endereco <= endereco*16 + 3;
           when '4' => endereco <= endereco*16 + 4;
           when '5' => endereco <= endereco*16 + 5;
           when '6' => endereco <= endereco*16 + 6;
           when '7' => endereco <= endereco*16 + 7;
           when '8' => endereco <= endereco*16 + 8;
           when '9' => endereco <= endereco*16 + 9;
           when 'A' => endereco <= endereco*16 + 10;
           when 'B' => endereco <= endereco*16 + 11;
           when 'C' => endereco <= endereco*16 + 12;
           when 'D' => endereco <= endereco*16 + 13;
           when 'E' => endereco <= endereco*16 + 14;
           when 'F' => endereco <= endereco*16 + 15;
           when others => null;
          end case;
                    
          case linha(4) is  
           when '0' => ops <= 0;
           when '1' => ops <= 1;
           when '2' => ops <= 2;
           when '3' => ops <= 3;
           when '4' => ops <= 4;
           when '5' => ops <= 5;
           when '6' => ops <= 6;
           when '7' => ops <= 7;
           when '8' => ops <= 8;
           when '9' => ops <= 9;
           when 'A' => ops <=10;
           when 'B' => ops <=11;
           when 'C' => ops <=12;
           when 'D' => ops <=13;
           when 'E' => ops <=14;
           when 'F' => ops <=15;
           when others => null;
          end case;
     
          wait for 1ps;
          
          case linha(5) is  
           when '0' => ops <= ops*16 + 0;
           when '1' => ops <= ops*16 + 1;
           when '2' => ops <= ops*16 + 2;
           when '3' => ops <= ops*16 + 3;
           when '4' => ops <= ops*16 + 4;
           when '5' => ops <= ops*16 + 5;
           when '6' => ops <= ops*16 + 6;
           when '7' => ops <= ops*16 + 7;
           when '8' => ops <= ops*16 + 8;
           when '9' => ops <= ops*16 + 9;
           when 'A' => ops <= ops*16 + 10;
           when 'B' => ops <= ops*16 + 11;
           when 'C' => ops <= ops*16 + 12;
           when 'D' => ops <= ops*16 + 13;
           when 'E' => ops <= ops*16 + 14;
           when 'F' => ops <= ops*16 + 15;
           when others => null;
          end case;
          
          wait for 1ps;
          
          go <= '1';    -- este sinal de go habilita a escrita na memoria
          
          wait for 1 ps;
          
          go <= '0';    -- tira o go
          
         end loop;

   end process;
                                     
end tb;
